/*
*****************************************************************************
* Copyright by ams AG                                                       *
* All rights are reserved.                                                  *
*                                                                           *
* IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
* THE SOFTWARE.                                                             *
*                                                                           *
* THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
* USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
* EXCLUDED.                                                                 *
*                                                                           *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
*****************************************************************************
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <linux/limits.h>

#include "mcu_tmf8701_config.h"
#include "tmf8701.h"
#include "tof_hex_interpreter.h"
#include "ams_linux_gpio.h"
#include "ams_linux_i2c.h"

#define EXP_PARAMS       1
#define GPIO_CE          (GPIO_16)
#define GPIO_IRQ         (GPIO_20)
#define TMF8701_I2C_ADDR (0x41)

static uint32_t exit_num_results = 0;
static struct tmf8701_ctx tmf8701 = {
    .i2c_addr = TMF8701_I2C_ADDR,
};

static int32_t do_fwdl(const char *filename)
{
    FILE *fwfile;
    size_t len = 0;
    ssize_t read = 0;
    char * line = NULL;
    int32_t error = 0;

    if (!filename)
        return -1;

    fwfile = fopen(filename, "r");

    if (!fwfile)
        return -1;

    error = tof8701_BL_upload_init(&tmf8701, &tmf8701.BL_app, BL_DEFAULT_SALT);
    if (error) {
        return error;
    }

    intelHexInterpreterInitialise();
    while ((read = getline(&line, &len, fwfile) != -1)) {
        error = intelHexHandleRecord(&tmf8701, &tmf8701.BL_app, len, line, 0);
        if (error)
            return error;
    }

    fclose(fwfile);
    if (line)
        free(line);

    return 0;
}

static int32_t read_factory_calibration_file(const char *filename, uint8_t*fac_cal, uint32_t size)
{
    FILE *fac_cal_file;
    char * line = NULL;
    size_t len = 0;
    ssize_t read = 0;
    uint32_t count = 0;
    int32_t error = 0;

    if (!filename || !fac_cal || size == 0)
        return -1;

    fac_cal_file = fopen(filename, "r");

    if (!fac_cal_file)
        return -1;

    while ((read = getline(&line, &len, fac_cal_file) != -1)) {
        error = sscanf(line, "%hhx", &fac_cal[count++]);
        if (error != 1)
            return -1;
        if (count == size)
            break; // we have filled the factory calibration buffer
    }

    printf("Factory calibration data: \n");
    for (uint32_t i = 0; i < size; i++) {
        printf("0x%02x ", fac_cal[i]);
    }
    printf("\n");

    fclose(fac_cal_file);
    if (line)
        free(line);

    return 0;
}

static int32_t start_factory_calib(struct tmf8701_ctx *tmf8701, const char *output_filename)
{
    int32_t flags = 0;
    int32_t error = 0;
    FILE *fac_cal_file;

    if (!tmf8701)
        return -1;

    fac_cal_file = fopen(output_filename, "w");

    if (!fac_cal_file)
        return -1;

    error = tof8701_app0_start_fac_calib(tmf8701);
    if (error) {
        fprintf(stderr, "Error starting factory calibration\n");
    }

    do {
        sleep(1);
        flags = tof8701_app0_interrupt_status(tmf8701);
        if (flags < 0) {
            fprintf(stderr, "Error reading int status\n");
            return -1;
        }
    } while (flags == 0);

    tof8701_app0_process_irq(tmf8701);

    for (int i = 0; i < APP0_FAC_CALIB_SIZE; i++) {
        fprintf(fac_cal_file, "0x%02x\n", tmf8701->app0_app.app0_fac_calib[i]);
    }
    fclose(fac_cal_file);
    printf("Factory calibration complete, saved to: '%s'\n", output_filename);

    return 0;
}

static void print_result(struct tmf8701_ctx *tmf8701)
{
    static int32_t last_num = 0;
    static int32_t cnt = 0;
    uint32_t num;
    uint32_t distance, conf;

    if (!tmf8701)
        return;

    num =tof8701_app0_get_last_result(tmf8701, &distance, &conf);
    if (num != last_num) {
        printf("Result num: %4u\tDistance: %6u mm\tconfidence: %4u\n",
               num, distance, conf);
        last_num = num;
        cnt++;
    }

    if (exit_num_results && cnt >= exit_num_results) {
        printf("Read %u results, exiting.\n", cnt);
        exit(write_gpio(GPIO_CE, 0));
    }

    return;
}

static void start_measurements(struct tmf8701_ctx *tmf8701, bool with_fac_cal, uint8_t *fac_cal, uint32_t len)
{
    int32_t rc = 0;

    if (!tmf8701) {
        fprintf(stderr, "Error null context pointer\n");
        exit(-1);
    }

    if (with_fac_cal) {
        if (!fac_cal || len == 0) {
            fprintf(stderr, "Error null fac_cal pointer, or size is 0\n");
        } else {
            rc = tof8701_app0_set_fac_calib(tmf8701, fac_cal, len);
            if (rc) {
                fprintf(stderr, "Error configuring factory calibration data: %d\n", rc);
                exit(rc);
            }
        }
    }

    (void) tof8701_app0_set_default_capture_settings(tmf8701);

    rc = tof8701_app0_start_capture(tmf8701);
    if (rc) {
        fprintf(stderr, "Error starting measurements: %d\n", rc);
        exit(rc);
    }

    while ( true ) {
        tof8701_app0_process_irq(tmf8701);
        print_result(tmf8701);
        usleep(5000); // poll period
    }
    return;
}

void catch_sigint(int parm)
{
    exit(write_gpio(GPIO_CE, 0));
}

int main(int argc, char * argv[])
{
    int c;
    uint32_t gpio_val = 0;
    int32_t rc = 0;
    char firmware_file[PATH_MAX] = "";
    bool do_factory_calib = false;
    bool fac_cal_supplied = false;
    char fac_cal_fname[PATH_MAX] = {0};
    uint8_t opt_fac_calib[APP0_FAC_CALIB_SIZE] = {0};

    signal(SIGINT, catch_sigint);

    while ((c = getopt (argc, argv, "hvz:c:n:")) != -1) {
        switch (c) {
            case 'h':
                printf("\n\n");
                printf("Help: ./%s [options] <firmware_file>\n",
                        MCU_TMF8701_PROJ_NAME);
                printf("\n\n");
                printf("\t<firmware_file>: The firmware patch to download to "
                       "the chip (required for tmf8701)\n");
                printf("\n\n");
                printf("Options:\n");
                printf("\t-h: display this help screen\n");
                printf("\t-c <input_fac_calib_file>: Factory calibration file "
                       "with newline-separated hexadecimal values\n");
                printf("\t-n <num_results>: read <num_results> and then exit\n");
                printf("\t-v: display version info\n");
                printf("\t-z <output_fac_calib_file>: perform factory "
                       " calibration and print to <output_fac_calib_file>\n");
                printf("\n\n");
                exit(0);
                break;
            case 'c':
                sscanf(optarg, "%s", fac_cal_fname);
                printf("Input factory calibration file: \'%s\'\n",
                       fac_cal_fname);
                rc = read_factory_calibration_file(fac_cal_fname,
                                                   opt_fac_calib,
                                                   sizeof(opt_fac_calib));
                if (rc) {
                    printf("Error reading Factory calibration file\n");
                    exit(rc);
                }
                fac_cal_supplied = true;
                break;
            case 'n':
                sscanf(optarg, "%u", &exit_num_results);
                break;
            case 'v':
                printf("mcu driver tmf8701 test tool: '%s' Version: %u.%u\n",
                        MCU_TMF8701_PROJ_NAME,
                        MCU_TMF8701_VERSION_MAJOR,
                        MCU_TMF8701_VERSION_MINOR);
                break;
            case 'z':
                sscanf(optarg, "%s", fac_cal_fname);
                printf("Output factory calibration file: \'%s\'\n",
                       fac_cal_fname);
                do_factory_calib = true;
                break;
            default:
                fprintf(stderr, "Error parsing cmdline args.\n");
                exit(1);
        }
    }

#ifdef _HAS_BUILTIN_FW
    if (argc - optind) {
        sscanf(argv[optind], "%s", firmware_file);
        printf("Firmware file: %s\n", firmware_file);
    }
#else
    if ((argc - optind) < EXP_PARAMS) {
        fprintf(stderr, "Error, expected \'%d\' parameters, "
                "\'%d\' received\n",
                EXP_PARAMS, (argc - optind));
        exit(1);
    }
    sscanf(argv[optind], "%s", firmware_file);
    printf("Firmware file: %s\n", firmware_file);
#endif

    // Init CHIP_EN GPIO
    rc = init_gpio(GPIO_CE,  OUT_DIR);
    if (rc) {
        fprintf(stderr, "Error initializing CE pin\n");
        exit(rc);
    }

    // Assert CHIP_EN GPIO
    rc = write_gpio(GPIO_CE, 1);
    if (rc) {
        fprintf(stderr, "Error asserting CE pin\n");
        exit(rc);
    }

    // Init INT GPIO
    /* Unused in this example program, we are polling the INTFLAG register */
    rc = init_gpio(GPIO_IRQ, IN_DIR);
    if (rc) {
        fprintf(stderr, "Error initializing IRQ pin\n");
        exit(rc);
    }

    // wait for i2c communication
    rc = tof8701_wait_for_cpu_ready(&tmf8701);
    if (rc) {
        fprintf(stderr, "Error waiting for CPU_READY status\n");
        exit(rc);
    }

    // initialize bootloader app
    tof8701_BL_init_app(&tmf8701.BL_app);

    // Download app0 firmware
    if (strlen(firmware_file)) {
        rc = do_fwdl(firmware_file);
        if (rc) {
            fprintf(stderr, "Error (%d) performing FWDL with file %s\n", rc,
                    firmware_file);
            exit(rc);
        }
    } else {
#ifdef _HAS_BUILTIN_FW
        printf("Using builtin fw image start addr: 0x%08x size: 0x%08x\n",
               tof_bin_image_start, tof_bin_image_length);
        rc = tof8701_BL_bin_download(&tmf8701, tof_bin_image_start,
                                     tof_bin_image, tof_bin_image_length);
        if (rc) {
            fprintf(stderr, "Error (%d) performing FWDL with built-in firmware\n", rc);
            exit(rc);
        }
#else
        fprintf(stderr, "Error, no firmware hex file given\n");
        exit(1);
#endif
    }

    // initialize app0 app
    tof8701_app0_init_app(&tmf8701);

    printf("App0 version: %u.%u.%u.%u\n",
            tmf8701.app0_app.version[0],tmf8701.app0_app.version[1], tmf8701.app0_app.version[2],
            tmf8701.app0_app.version[3] | (tmf8701.app0_app.version[4] << 8));

    if (do_factory_calib) {
        rc = start_factory_calib(&tmf8701, fac_cal_fname);
        if (rc != 0) {
            fprintf(stderr, "Error performing factory calibration: %d\n", rc);
        }
        exit(rc);
    } else {
        start_measurements(&tmf8701, fac_cal_supplied,
                           opt_fac_calib, sizeof(opt_fac_calib)); // this function never returns
    }

    exit(0);
}

