extern "C" {
#include "ams_linux_gpio.c"
}
#include <gtest/gtest.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <fff.h>

FAKE_VALUE_FUNC(int, system, const char *);
FAKE_VALUE_FUNC(FILE *, popen, const char *, const char *);
FAKE_VALUE_FUNC(int, pclose, FILE *);
FAKE_VALUE_FUNC_VARARG(int, fscanf, FILE *, const char *,...);

#define FFF_FAKES_LIST(FAKE)  \
    FAKE(system) \
    FAKE(fscanf) \
    FAKE(popen)

class AmsLinuxGpioTest : public testing::Test
{
    void SetUp() {
        FFF_FAKES_LIST(RESET_FAKE);
        FFF_RESET_HISTORY();
    }
};

static char sys_arg[PATH_MAX] ={0};
static int sys_rc = 0;;
int sys_fake (const char *s) {
    strcpy(sys_arg, s);
    return sys_rc;
}

static FILE * popen_rc = 0;
static char popen_arg[PATH_MAX] ={0};
FILE * p_fake (const char *s, const char *m) {
    strcpy(popen_arg, s);
    return popen_rc;
}

TEST_F (AmsLinuxGpioTest, test_gpio_is_init)
{
    bool rc = false;
    char arg[PATH_MAX] = {0};
    system_fake.custom_fake = sys_fake;

    rc = gpio_is_init(0);
    ASSERT_EQ(rc, false);

    sys_rc = 1;
    rc = gpio_is_init(1);
    ASSERT_EQ(rc, false);

    sys_rc = 0;
    snprintf(arg, PATH_MAX, "%s%u", sysfs_gpio, 1);
    rc = gpio_is_init(1);
    ASSERT_EQ(rc, true);
    ASSERT_STRNE((char *)NULL, strstr(sys_arg, arg));
};

TEST_F (AmsLinuxGpioTest, test_init_gpio)
{
    int32_t rc = 0;
    char arg[PATH_MAX] = {0};

    rc = init_gpio(0, IN_DIR);
    ASSERT_LT(rc, 0);

    rc = init_gpio(1, 5);
    ASSERT_LT(rc, 0);

    // test not init happy path
    int vals[] = {1, 0};
    SET_RETURN_SEQ(system, vals, 2);
    rc = init_gpio(1, OUT_DIR);
    ASSERT_EQ(rc, 0);

    system_fake.custom_fake = sys_fake;

    // test already init happy path
    sys_rc = 0;
    rc = init_gpio(1, OUT_DIR);
    ASSERT_EQ(rc, 0);
    snprintf(arg, PATH_MAX, "echo out > %s%u", sysfs_gpio, 1);
    ASSERT_STRNE((char *)NULL, strstr(sys_arg, arg));

    // test not init fail export
    sys_rc = 1;
    rc = init_gpio(1, OUT_DIR);
    ASSERT_EQ(rc, 1);
    snprintf(arg, PATH_MAX, "echo %u > %s", 1, sysfs_export);
    ASSERT_STRNE((char *)NULL, strstr(sys_arg, arg));

    rc = init_gpio(1, OUT_DIR);
    ASSERT_EQ(rc, 1);

};

TEST_F (AmsLinuxGpioTest, test_deinit_gpio)
{
    int32_t rc = 0;
    char arg[PATH_MAX] = {0};

    // bad gpio
    rc = deinit_gpio(0);
    ASSERT_LT(rc, 0);

    // not init
    system_fake.return_val = 1;
    rc = deinit_gpio(1);
    ASSERT_LT(rc, 0);

    // test failed export
    int vals[] = {0, -1};
    SET_RETURN_SEQ(system, vals, 2);
    rc = deinit_gpio(1);
    ASSERT_LT(rc, 0);

    // test happy path deinit
    system_fake.custom_fake = sys_fake;
    sys_rc = 0;
    rc = deinit_gpio(1);
    ASSERT_EQ(rc, 0);
    snprintf(arg, PATH_MAX, "%u > %s", 1, sysfs_unexport);
    ASSERT_STRNE((char *)NULL, strstr(sys_arg, arg));
}

TEST_F (AmsLinuxGpioTest, test_read_gpio)
{
    int32_t rc = 0;
    char arg[PATH_MAX] = {0};
    uint32_t val = 0;

    // bad gpio
    rc = read_gpio(0, &val);
    ASSERT_LT(rc, 0);

    // bad param
    rc = read_gpio(0, NULL);
    ASSERT_LT(rc, 0);

    // bad FILE *
    popen_fake.return_val = 0;
    rc = read_gpio(1, &val);
    ASSERT_LT(rc, 0);

    // happy path
    popen_rc = (FILE *)1;
    popen_fake.custom_fake = p_fake;
    pclose_fake.return_val = 0;
    fscanf_fake.return_val = 0;
    rc = read_gpio(1, &val);
    ASSERT_EQ(rc, 0);
    snprintf(arg, PATH_MAX, "%s%u/value", sysfs_gpio, 1);
    ASSERT_STRNE((char *)NULL, strstr(popen_arg, arg));
}

TEST_F (AmsLinuxGpioTest, test_write_gpio)
{
    int32_t rc = 0;
    char arg[PATH_MAX] = {0};
    uint32_t val = 0;

    // bad gpio
    rc = write_gpio(0, val);
    ASSERT_LT(rc, 0);

    // happy path
    sys_rc = 0;
    system_fake.custom_fake = sys_fake;
    rc = write_gpio(1, val);
    ASSERT_EQ(rc, 0);
    snprintf(arg, PATH_MAX, "echo %u > %s%u/value",val, sysfs_gpio, 1);
    ASSERT_STRNE((char *)NULL, strstr(sys_arg, arg));
}
