extern "C" {
#include "tmf8701.c"
}
#include <gtest/gtest.h>
#include <fff.h>

FAKE_VALUE_FUNC(int32_t, read_i2c_block, uint32_t,uint8_t,uint8_t*,uint32_t);
FAKE_VALUE_FUNC(int32_t, write_i2c_block, uint32_t,uint8_t,const uint8_t*,uint32_t);
FAKE_VALUE_FUNC(int, clock_gettime, clockid_t,struct timespec *);

#define FFF_FAKES_LIST(FAKE)           \
    FAKE(clock_gettime)               \
    FAKE(read_i2c_block)               \
    FAKE(write_i2c_block)

static uint32_t read_idx = 0;
static uint32_t write_idx = 0;
static int32_t r_rcbuf[10] = {0};
static int32_t w_rcbuf[10] = {0};
static int32_t * const read_rc = r_rcbuf;
static int32_t * const write_rc = w_rcbuf;
static uint8_t fakebuf[MAX_REGS] = {0};
static int32_t fake_read(uint32_t,uint8_t r,uint8_t *b,uint32_t l) {
    int32_t rc = 0;
    memcpy(b, &fakebuf[r], l);
    rc = r_rcbuf[read_idx];
    read_idx = (read_idx + 1) % 10;
    return rc;
}
static int32_t fake_write(uint32_t,uint8_t r,const uint8_t *b,uint32_t l) {
    int32_t rc = 0;
    memcpy( &fakebuf[r], b, l);
    rc = w_rcbuf[write_idx];
    write_idx = (write_idx + 1) % 10;
    return rc;
}
static int32_t fake_blwrite(uint32_t,uint8_t r,const uint8_t *b,uint32_t l) {
    int32_t rc = 0;
    memcpy( &fakebuf[r], b, l);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_BL_CMD_STAT+1] = 0x02;
    fakebuf[REG_BL_CMD_STAT+2] = 0x04;
    fakebuf[REG_BL_CMD_STAT+2] = 0x08;
    fakebuf[REG_BL_CMD_STAT+3] = 0xF0;
    rc = w_rcbuf[write_idx];
    write_idx = (write_idx + 1) % 10;
    return rc;
}
static int gettime_rc = 0;
static struct timespec fake_ts = {0};
static int fake_gettime(clockid_t, struct timespec *ts) {
    memcpy(ts, &fake_ts, sizeof(fake_ts));
    return gettime_rc;
}

class TMF8701Test : public testing::Test
{
    void SetUp() {
        FFF_FAKES_LIST(RESET_FAKE);
        FFF_RESET_HISTORY();
        read_idx = 0;
        write_idx = 0;
        memset(r_rcbuf, 0, 10);
        memset(w_rcbuf, 0, 10);
        memset(fakebuf, 0, MAX_REGS);
    }
};

TEST_F (TMF8701Test, test_is_BL_cmd_busy)
{
    int32_t rc;
    struct tmf8701_ctx test_ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    rc = is_BL_cmd_busy(&test_ctx);
    ASSERT_EQ(0, rc);
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = is_BL_cmd_busy(&test_ctx);
    ASSERT_EQ(1, rc);
}

TEST_F (TMF8701Test, test_TMF8701_cr_addpair)
{
    uint32_t h;
    uint32_t d;
    TMF8701_cr cr = {0};
    TMF8701_cr_init(&cr);
    h = 5;
    d = 0;
    TMF8701_cr_addpair(&cr, h, d);
    h = 1;
    d = 5;
    TMF8701_cr_addpair(&cr, h, d);
    ASSERT_EQ(cr.count, 1);
    h = 2;
    d = 10;
    TMF8701_cr_addpair(&cr, h, d);
    h = 3;
    d = 15;
    TMF8701_cr_addpair(&cr, h, d);
    ASSERT_EQ(cr.trim_count, 1);
    ASSERT_EQ(cr.ratioQ15, 6554);
    ASSERT_GE(cr.count, 3);
    h = 999;
    d = 15;
    TMF8701_cr_addpair(&cr, h, d);
    ASSERT_GE(cr.count, 0);
    h = 1;
    d = 15;
    TMF8701_cr_addpair(&cr, h, d);
    ASSERT_EQ(cr.count, 1);
    h = 1;
    d = 5;
    TMF8701_cr_addpair(&cr, h, d);
    h = 1;
    d = 4;
    TMF8701_cr_addpair(&cr, h, d);
    ASSERT_EQ(cr.count, 1);
};

TEST_F (TMF8701Test, test_app0_read_fac_calib)
{
    int32_t rc = 0;
    read_i2c_block_fake.custom_fake = fake_read;
    *read_rc = 0;
    struct tmf8701_ctx ctx = {0};
    for (int i = REG_APP0_DATA_START; i < (REG_APP0_DATA_START + APP0_FAC_CALIB_SIZE); i++) {
        fakebuf[i] = i;
    }
    rc = tof8701_app0_read_fac_calib(&ctx);
    ASSERT_EQ(rc, 0);
    for (int i = REG_APP0_DATA_START, j = 0; j < APP0_FAC_CALIB_SIZE; i++,j++) {
        ASSERT_EQ(fakebuf[i], ctx.app0_app.app0_fac_calib[j]);
    }
};

TEST_F (TMF8701Test, test_app0_write_cal_data)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    ctx.app0_app.cal_flags = APP0_FAC_CAL_FLAG | APP0_ALG_STATE_FLAG | APP0_CFG_DATA_FLAG;
    int32_t write_vals[] = {-1, 0, -1, 0, 0, -1};
    SET_RETURN_SEQ(write_i2c_block, write_vals, sizeof(write_vals)/sizeof(write_vals[0]));
    rc = tof8701_app0_write_cal_data(&ctx);
    ASSERT_LT(rc, 0);
    rc = tof8701_app0_write_cal_data(&ctx);
    ASSERT_LT(rc, 0);
    rc = tof8701_app0_write_cal_data(&ctx);
    ASSERT_LT(rc, 0);
    write_i2c_block_fake.custom_fake = fake_write;
    *write_rc = 0;
    for (int i = REG_APP0_DATA_START;
         i < (REG_APP0_DATA_START + APP0_FAC_CALIB_SIZE + APP0_ALG_STATE_SIZE + APP0_CFG_DATA_SIZE); i++) {
        for (int j = 0; j < APP0_FAC_CALIB_SIZE; i++,j++) {
            ctx.app0_app.app0_fac_calib[j] = i;
        }
        for (int j = 0; j < APP0_ALG_STATE_SIZE; i++,j++) {
            ctx.app0_app.result.alg_state[j] = i;
        }
        for (int j = 0; j < APP0_CFG_DATA_SIZE; i++,j++) {
            ctx.app0_app.app0_cfg_data[j] = i;
        }
    }
    rc = tof8701_app0_write_cal_data(&ctx);
    ASSERT_EQ(rc, 0);
    for (int i = REG_APP0_DATA_START; i < (REG_APP0_DATA_START + APP0_FAC_CALIB_SIZE + APP0_ALG_STATE_SIZE + APP0_CFG_DATA_SIZE); i++) {
        for (int j = 0; j < APP0_FAC_CALIB_SIZE; i++,j++) {
            ASSERT_EQ(fakebuf[i], ctx.app0_app.app0_fac_calib[j]);
        }
        for (int j = 0; j < APP0_ALG_STATE_SIZE; i++,j++) {
            ASSERT_EQ(fakebuf[i], ctx.app0_app.result.alg_state[j]);
        }
        for (int j = 0; j < APP0_CFG_DATA_SIZE; i++,j++) {
            ASSERT_EQ(fakebuf[i], ctx.app0_app.app0_cfg_data[j]);
        }
    }
};

TEST_F (TMF8701Test, test_get_usec_timestamp)
{
    int32_t rc = 0;
    uint32_t val = 0;
    rc = get_usec_timestamp(NULL);
    ASSERT_LT(rc, 0);
    rc = get_usec_timestamp(&val);
    ASSERT_EQ(rc, 0);
    ASSERT_NE(val, 0);
}

TEST_F (TMF8701Test, test_TMF8701_cr_map)
{
    TMF8701_cr cr = {0};
    uint32_t in_dis = 100;
    uint32_t out_dis;
    cr.ratioQ15 = 6554; // test target ratio of 0.2
    out_dis = TMF8701_cr_map(&cr, in_dis);
    ASSERT_EQ(in_dis, out_dis);
}

TEST_F (TMF8701Test, test_tof8701_app0_osc_trim)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};

    int32_t w_vals[10];
    int32_t r_vals[10];

    w_vals[0] = -1;
    r_vals[0] = 0;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 1);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 1);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    w_vals[0] = 0;
    r_vals[0] = -1;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 1);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 1);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    w_vals[0] = 0;
    r_vals[0] = 0;
    r_vals[1] = 0;
    r_vals[2] = -1;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 1);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 3);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    w_vals[0] = 0;
    r_vals[0] = 0;
    r_vals[1] = 0;
    r_vals[2] = 0;
    r_vals[3] = 0;
    r_vals[4] = 0;
    r_vals[5] = -1;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 1);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 6);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    w_vals[0] = 0;
    w_vals[1] = 0;
    w_vals[2] = -1;
    r_vals[0] = 0;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 3);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 1);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    w_vals[0] = 0;
    w_vals[1] = 0;
    w_vals[2] = 0;
    w_vals[3] = 0;
    w_vals[4] = 0;
    w_vals[5] = -1;
    r_vals[0] = 0;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 6);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 1);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    w_vals[0] = 0;
    w_vals[1] = 0;
    w_vals[2] = 0;
    w_vals[3] = 0;
    w_vals[4] = 0;
    w_vals[5] = 0;
    w_vals[6] = 0;
    w_vals[7] = 0;
    w_vals[8] = 0;
    w_vals[9] = -1;
    r_vals[0] = 0;
    SET_RETURN_SEQ(write_i2c_block, w_vals, 10);
    SET_RETURN_SEQ(read_i2c_block, r_vals, 1);
    rc = tof8701_app0_osc_trim(&ctx, 1);
    ASSERT_LT(rc, 0);

    write_i2c_block_fake.custom_fake = fake_write;
    *write_rc = 0;
    read_i2c_block_fake.custom_fake = fake_read;
    *read_rc = 0;
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    fakebuf[3] = 0xFF;
    fakebuf[6] &= ~(1 << 6);
    rc = tof8701_app0_osc_trim(&ctx, -1);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[3], 0xFE);
    ASSERT_EQ(((fakebuf[6] >> 6) & 0x01), 1);
}

TEST_F (TMF8701Test, test_app0_osc_trim_up)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    int32_t write_vals[] = {-1, 0};
    SET_RETURN_SEQ(write_i2c_block, write_vals, sizeof(write_vals)/sizeof(write_vals[0]));
    int32_t read_vals[] = {0};
    SET_RETURN_SEQ(read_i2c_block, read_vals, sizeof(read_vals)/sizeof(read_vals[0]));
    rc = tof8701_app0_osc_trim_up(&ctx);
    ASSERT_LT(rc, 0);

    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    write_i2c_block_fake.custom_fake = fake_write;
    w_rcbuf[0] = 0;
    w_rcbuf[1] = 0;
    w_rcbuf[2] = -1;
    read_i2c_block_fake.custom_fake = fake_read;
    rc = tof8701_app0_osc_trim_up(&ctx);
    ASSERT_LT(rc, 0);
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    write_i2c_block_fake.custom_fake = fake_write;
    *write_rc = 0;
    read_i2c_block_fake.custom_fake = fake_read;
    *read_rc = 0;
    rc = tof8701_app0_osc_trim_up(&ctx);
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, test_app0_osc_trim_down)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    int32_t write_vals[] = {-1, 0};
    SET_RETURN_SEQ(write_i2c_block, write_vals, sizeof(write_vals)/sizeof(write_vals[0]));
    int32_t read_vals[] = {0};
    SET_RETURN_SEQ(read_i2c_block, read_vals, sizeof(read_vals)/sizeof(read_vals[0]));
    rc = tof8701_app0_osc_trim_down(&ctx);
    ASSERT_LT(rc, 0);

    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    write_i2c_block_fake.custom_fake = fake_write;
    w_rcbuf[0] = 0;
    w_rcbuf[1] = 0;
    w_rcbuf[2] = -1;
    read_i2c_block_fake.custom_fake = fake_read;
    rc = tof8701_app0_osc_trim_down(&ctx);
    ASSERT_LT(rc, 0);
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    write_i2c_block_fake.custom_fake = fake_write;
    *write_rc = 0;
    read_i2c_block_fake.custom_fake = fake_read;
    *read_rc = 0;
    rc = tof8701_app0_osc_trim_down(&ctx);
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_app0_check_osc)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    ctx.app0_app.clk_trim_enable = true;
    ctx.app0_app.clk_cr.trim_count = APP0_DEFAULT_CLK_ITERATIONS; // trigger osc trim
    ctx.app0_app.clk_cr.ratioQ15 = (APP0_FREQ_RATIO_HITH_Q15 + APP0_FREQ_RATIO_LOTH_Q15) / 2;
    rc = tof8701_app0_check_osc(&ctx);
    ASSERT_EQ(rc, 0);
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    write_i2c_block_fake.custom_fake = fake_write;
    *write_rc = -1;
    read_i2c_block_fake.custom_fake = fake_read;
    *read_rc = -1;
    ctx.app0_app.clk_cr.ratioQ15 = APP0_FREQ_RATIO_HITH_Q15;
    ctx.app0_app.clk_cr.trim_count = APP0_DEFAULT_CLK_ITERATIONS; // trigger osc trim
    rc = tof8701_app0_check_osc(&ctx);
    ASSERT_LT(rc, 0);
    ctx.app0_app.clk_cr.ratioQ15 = APP0_FREQ_RATIO_LOTH_Q15;
    ctx.app0_app.clk_cr.trim_count = APP0_DEFAULT_CLK_ITERATIONS; // trigger osc trim
    rc = tof8701_app0_check_osc(&ctx);
    ASSERT_LT(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_app0_clock_skew_correct_results)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    fakebuf[REG_CLK_STATUS] = 0x0;
    ctx.app0_app.clk_cr.ratioQ15 = APP0_EXP_FREQ_RATIO_Q15;
    ctx.app0_app.result.distance_mm[0] = 300 & 0xFF;
    ctx.app0_app.result.distance_mm[1] = 300 >> 8;
    rc = tof8701_app0_clock_skew_correct_results(&ctx);
    ASSERT_EQ(rc, 0);
    int dist = TMF8701_cr_map(&ctx.app0_app.clk_cr, 300);
    ASSERT_EQ((dist & 0xFF), ctx.app0_app.result.distance_mm[0]);
    ASSERT_EQ((dist >> 8), ctx.app0_app.result.distance_mm[1]);
    ctx.app0_app.clk_cr.ratioQ15 = APP0_EXP_FREQ_RATIO_Q15;
    ctx.app0_app.clk_cr.count = APP0_RATIO_MINCOUNT + 1;
    ctx.app0_app.result.distance_mm[0] = 300 & 0xFF;
    ctx.app0_app.result.distance_mm[1] = 300 >> 8;
    clock_gettime_fake.custom_fake = fake_gettime;
    gettime_rc = 0;
    fake_ts.tv_sec = 999;
    rc = tof8701_app0_clock_skew_correct_results(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(ctx.app0_app.clk_cr.count, 0);
    ctx.app0_app.clk_cr.ratioQ15 = APP0_EXP_FREQ_RATIO_Q15;
    ctx.app0_app.result.distance_mm[0] = 700 & 0xFF;
    ctx.app0_app.result.distance_mm[1] = 700 >> 8;
    rc = tof8701_app0_clock_skew_correct_results(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(0, ctx.app0_app.result.distance_mm[0]);
    ASSERT_EQ(0, ctx.app0_app.result.distance_mm[1]);
    ASSERT_EQ(ctx.app0_app.result.info & APP0_RESULT_CONF_MASK, 0);
    clock_gettime_fake.custom_fake = fake_gettime;
    gettime_rc = -1;
    rc = tof8701_app0_clock_skew_correct_results(&ctx);
    ASSERT_LT(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_app0_read_meas_result)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.return_val = -1;
    rc = tof8701_app0_read_meas_result(&ctx);
    ASSERT_LT(rc, 0);
    for (int i = REG_APP0_DATA_START; i < APP0_RESULT_SIZE; i++) {
        fakebuf[i] = i;
    }
    ctx.app0_app.status.contents = MEAS_RESULT_CONTENTS_ID;
    read_i2c_block_fake.custom_fake = fake_read;
    *read_rc = 0;
    rc = tof8701_app0_read_meas_result(&ctx);
    for (int i = REG_APP0_DATA_START, j = 0; i < APP0_RESULT_SIZE; j++, i++) {
        ASSERT_EQ(fakebuf[i],ctx.app0_app.result.buf[j]);
    }
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, test_wait_for_cpu_ready)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    rc = tof8701_wait_for_cpu_ready(NULL);
    ASSERT_LT(rc, 0);
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATE] = 0x1;
    rc = tof8701_wait_for_cpu_ready(&ctx);
    ASSERT_EQ(rc, 0);
    fakebuf[REG_CPU_STATUS] = 0x40;
    fakebuf[REG_APP0_STATE] = 0x1;
    rc = tof8701_wait_for_cpu_ready(&ctx);
    ASSERT_EQ(rc, 0);
    fakebuf[REG_CPU_STATUS] = 0;
    *write_rc = -1;
    rc = tof8701_wait_for_cpu_ready(&ctx);
    ASSERT_LT(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_goto_standby)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    rc = tof8701_goto_standby(NULL);
    ASSERT_LT(rc, 0);
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_CPU_STATUS] = 0x41;
    rc = tof8701_goto_standby(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_CPU_STATUS], 0x40);
}

TEST_F (TMF8701Test, test_tof8701_wake_from_standby)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    rc = tof8701_wake_from_standby(NULL);
    ASSERT_LT(rc, 0);
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_CPU_STATUS] = 0x40;
    rc = tof8701_wake_from_standby(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_CPU_STATUS], 0x41);
}

TEST_F (TMF8701Test, test_dump_i2c_regs)
{
    struct tmf8701_ctx ctx = {0};
    for (int i = 0; i < MAX_REGS; i++) {
        fakebuf[i] = i;
    }
    read_i2c_block_fake.custom_fake = fake_read;
    tof8701_dump_i2c_regs(NULL, 253, 255);
    tof8701_dump_i2c_regs(&ctx, 253, 255);
}

TEST_F (TMF8701Test, verify_BL_init_called)
{
    struct tof8701_BL_application test_BL = {0};
    tof8701_BL_init_app(&test_BL);
    ASSERT_EQ(test_BL.app_id, APP_ID_BOOTLOADER);
};

TEST_F (TMF8701Test, test_BL_read_status)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    *read_rc = -1;
    rc = tof8701_BL_read_status(&ctx, &ctx.BL_app, 2);
    ASSERT_LT(rc, 0);
    *read_rc = 0;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_read_status(&ctx, &ctx.BL_app, 2);
    ASSERT_LT(rc, 0);
    for (int i = 0; i < MAX_REGS; i++) {
        fakebuf[i] = i;
    }
    rc = tof8701_BL_read_status(&ctx, &ctx.BL_app, 1);
    ASSERT_LT(rc, 0);
    read_idx = 0;
    r_rcbuf[0] = 0;
    r_rcbuf[1] = -1;
    rc = tof8701_BL_read_status(&ctx, &ctx.BL_app, 1);
    ASSERT_LT(rc, 0);
    r_rcbuf[1] = 0;
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_BL_CMD_STAT+1] = 0;
    fakebuf[REG_BL_CMD_STAT+2] = 0xFE;
    rc = tof8701_BL_read_status(&ctx, &ctx.BL_app, 1);
    ASSERT_EQ(rc, 0);
};

TEST_F (TMF8701Test, test_tof8701_calc_chksum)
{
    uint8_t testcksum = 0;
    fakebuf[REG_BL_CMD_STAT] = 0x10;
    fakebuf[REG_BL_CMD_STAT+1] = 0;
    testcksum = tof8701_calc_chksum(&fakebuf[REG_BL_CMD_STAT], 2);
    ASSERT_EQ(testcksum, 0xEF);
}

TEST_F (TMF8701Test, test_tof8701_BL_send_rcv_cmd)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_send_rcv_cmd(&ctx, &ctx.BL_app);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_BL_CMD_STAT+1] = 0;
    fakebuf[REG_BL_CMD_STAT+2] = 0xFE;
    rc = tof8701_BL_send_rcv_cmd(&ctx, &ctx.BL_app);
    ASSERT_EQ(rc, 0);
    write_idx = 0;
    *write_rc = -1;
    rc = tof8701_BL_send_rcv_cmd(&ctx, &ctx.BL_app);
    ASSERT_LT(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_BL_send_cmd)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_send_cmd(&ctx, &ctx.BL_app);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    ctx.BL_app.BL_command.short_cmd.command = 0x10;
    rc = tof8701_BL_send_cmd(&ctx, &ctx.BL_app);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(0x10, fakebuf[REG_BL_CMD_STAT]);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    write_idx = 0;
    *write_rc = -1;
    rc = tof8701_BL_send_cmd(&ctx, &ctx.BL_app);
    ASSERT_LT(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_BL_short_cmd)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_short_cmd(&ctx, &ctx.BL_app, BL_RESET);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_BL_CMD_STAT+1] = 0;
    fakebuf[REG_BL_CMD_STAT+2] = 0xFE;
    rc = tof8701_BL_short_cmd(&ctx, &ctx.BL_app, (tof8701_bootloader_cmd)0x01);
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_BL_reset)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_reset(&ctx, &ctx.BL_app);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_CPU_STATUS] = 0x41;
    rc = tof8701_BL_reset(&ctx, &ctx.BL_app);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_BL_CMD_STAT], BL_RESET);
}

TEST_F (TMF8701Test, test_tof8701_BL_ram_remap)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_ram_remap(&ctx, &ctx.BL_app);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_CPU_STATUS] = 0x41;
    rc = tof8701_BL_ram_remap(&ctx, &ctx.BL_app);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_BL_CMD_STAT], BL_RAMREMAP_RESET);
}

TEST_F (TMF8701Test, test_tof8701_BL_addr_ram)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_addr_ram(&ctx, &ctx.BL_app, 0);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_CPU_STATUS] = 0x41;
    write_i2c_block_fake.custom_fake = fake_blwrite;
    rc = tof8701_BL_addr_ram(&ctx, &ctx.BL_app, 0);
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_BL_read_ram)
{
    int32_t rc = 0;
    uint8_t val[255];
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_read_ram(&ctx, &ctx.BL_app, val, sizeof(val));
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_CPU_STATUS] = 0x41;
    write_i2c_block_fake.custom_fake = fake_blwrite;
    rc = tof8701_BL_read_ram(&ctx, &ctx.BL_app, val, sizeof(val));
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_BL_CMD_STAT+2], val[0]);
    ASSERT_EQ(fakebuf[REG_BL_CMD_STAT+3], val[1]);
}

TEST_F (TMF8701Test, test_tof8701_BL_write_ram)
{
    int32_t rc = 0;
    uint8_t val[255];
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_write_ram(&ctx, &ctx.BL_app, val, sizeof(val));
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_CPU_STATUS] = 0x41;
    write_i2c_block_fake.custom_fake = fake_blwrite;
    rc = tof8701_BL_write_ram(&ctx, &ctx.BL_app, val, sizeof(val));
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_BL_upload_init)
{
    int32_t rc = 0;
    uint8_t val = 0x29;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_BL_CMD_STAT] = 0xFF;
    rc = tof8701_BL_upload_init(&ctx, &ctx.BL_app, val);
    ASSERT_LT(rc, 0);
    fakebuf[REG_BL_CMD_STAT] = 0x01;
    fakebuf[REG_CPU_STATUS] = 0x41;
    rc = tof8701_BL_upload_init(&ctx, &ctx.BL_app, val);
    ASSERT_LT(rc, 0);
    ASSERT_EQ(BL_UPLOAD_INIT, fakebuf[REG_BL_CMD_STAT]);
    ASSERT_EQ(val, fakebuf[REG_BL_CMD_STAT+2]);
}

TEST_F (TMF8701Test, test_tof8701_BL_bin_download)
{
    int32_t rc = 0;
    uint8_t val[] = {1, 2};
    struct tmf8701_ctx ctx = {0};
    rc = tof8701_BL_bin_download(NULL, 0, val, sizeof(val));
    ASSERT_LT(rc, 0);
    fakebuf[REG_CPU_STATUS] = 0x41;
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_blwrite;
    *write_rc = -1;
    rc = tof8701_BL_bin_download(&ctx, 0, val, sizeof(val));
    ASSERT_LT(rc, 0);
    write_idx = 0;
    w_rcbuf[0] = 0;
    w_rcbuf[1] = -1;
    rc = tof8701_BL_bin_download(&ctx, 0, val, sizeof(val));
    ASSERT_LT(rc, 0);
    write_idx = 0;
    w_rcbuf[0] = 0;
    w_rcbuf[1] = 0;
    w_rcbuf[2] = -1;
    rc = tof8701_BL_bin_download(&ctx, 0, val, sizeof(val));
    ASSERT_LT(rc, 0);
    write_idx = 0;
    w_rcbuf[0] = 0;
    w_rcbuf[1] = 0;
    w_rcbuf[2] = 0;
    rc = tof8701_BL_bin_download(&ctx, 0, val, sizeof(val));
    ASSERT_EQ(rc, 0);
}

TEST_F (TMF8701Test, verify_app0_init_called)
{
    int32_t rc = 0;

    rc = tof8701_app0_init_app(NULL);
    ASSERT_EQ(rc, -1);

    struct tmf8701_ctx test_ctx = {0};
    read_i2c_block_fake.return_val = 0;
    rc = tof8701_app0_init_app(&test_ctx);
    ASSERT_EQ(test_ctx.app0_app.app_id, APP_ID_APP0);
    ASSERT_EQ(rc, 0);

    read_i2c_block_fake.return_val = 1;
    rc = tof8701_app0_init_app(&test_ctx);
    ASSERT_NE(rc, 0);

    read_i2c_block_fake.return_val = 0;
    write_i2c_block_fake.return_val = 1;
    rc = tof8701_app0_init_app(&test_ctx);
    ASSERT_NE(rc, 0);
};

TEST_F (TMF8701Test, test_tof8701_app0_wait_for_idle)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_wait_for_idle(NULL, 1000);
    ASSERT_EQ(rc, -1);
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATUS] = 0x1;
    fakebuf[REG_APP0_STATE] = 0xFF;
    rc = tof8701_app0_wait_for_idle(&ctx, 1000);
    ASSERT_EQ(rc, -1);
    fakebuf[REG_APP0_STATE] = APP0_IDLE_STATE;
    rc = tof8701_app0_wait_for_idle(&ctx, 1000);
    ASSERT_EQ(rc, 0);
}

static int32_t fake_majver_read(uint32_t,uint8_t,uint8_t*b,uint32_t l)
{
    if (l != 1) return -1;
    b[0] = 1;
        return 0;
}

static int32_t fake_minver_read(uint32_t,uint8_t,uint8_t*b,uint32_t l)
{
    if (l != 4) return -1;
    b[0] = 2;
    b[1] = 3;
    b[2] = 4;
    b[3] = 5;
    return 0;
}

TEST_F (TMF8701Test, test_app0_read_version)
{
    int32_t rc = 0;
    rc = tof8701_app0_read_version(NULL);
    ASSERT_EQ(rc, -1);
    int32_t return_vals[] = {0, -1, -1, 0, 0};
    struct tmf8701_ctx test_ctx = {0};
    SET_RETURN_SEQ(read_i2c_block, return_vals, sizeof(return_vals)/sizeof(return_vals[0]));
    rc = tof8701_app0_read_version(&test_ctx);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_read_version(&test_ctx);
    ASSERT_EQ(rc, -1);
    int32_t (*custom_fakes[])(uint32_t,uint8_t,uint8_t*,uint32_t) =
        { fake_majver_read, fake_minver_read };
    SET_CUSTOM_FAKE_SEQ(read_i2c_block, custom_fakes, sizeof(custom_fakes)/sizeof(custom_fakes[0]));
    rc = tof8701_app0_read_version(&test_ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(test_ctx.app0_app.version[0], 1);
    ASSERT_EQ(test_ctx.app0_app.version[1], 2);
    ASSERT_EQ(test_ctx.app0_app.version[2], 3);
    ASSERT_EQ(test_ctx.app0_app.version[3], 4);
    ASSERT_EQ(test_ctx.app0_app.version[4], 5);
}

TEST_F (TMF8701Test, test_tof8701_app0_enable_int)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_enable_int(NULL);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_enable_int(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_INT_ENABLE], IRQ_RESULTS | IRQ_ERROR);
}

TEST_F (TMF8701Test, test_tof8701_app0_start_fac_calib)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_start_fac_calib(NULL);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_start_fac_calib(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_APP0_CMD], APP0_CMD_FAC_CALIB);
}

TEST_F (TMF8701Test, test_tof8701_app0_set_default_capture_settings)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_set_default_capture_settings(NULL);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_set_default_capture_settings(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(ctx.app0_app.capture_cmd.iterations[0], APP0_ITERATIONS_DEFAULT & 0xFF);
    ASSERT_EQ(ctx.app0_app.capture_cmd.iterations[1], APP0_ITERATIONS_DEFAULT >> 8);
}

TEST_F (TMF8701Test, test_tof8701_app0_start_capture)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_start_capture(NULL);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_start_capture(&ctx);
    ASSERT_EQ(rc, 0);
    write_idx = 0;
    *write_rc = -1;
    ctx.app0_app.cal_flags |= APP0_FAC_CAL_FLAG;
    rc = tof8701_app0_start_capture(&ctx);
    ASSERT_LT(rc, 0);
}

TEST_F (TMF8701Test, test_tof8701_app0_stop_capture)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_stop_capture(NULL);
    ASSERT_EQ(rc, -1);
    write_idx = 0;
    *write_rc = -1;
    rc = tof8701_app0_stop_capture(&ctx);
    ASSERT_LT(rc, 0);
    w_rcbuf[1] = 0;
    w_rcbuf[2] = -1;
    rc = tof8701_app0_stop_capture(&ctx);
    ASSERT_LT(rc, 0);
    write_idx = 0;
    *write_rc = 0;
    fakebuf[REG_CPU_STATUS] = 0x41;
    fakebuf[REG_APP0_STATUS] = 0x1;
    fakebuf[REG_APP0_STATE] = APP0_IDLE_STATE;
    rc = tof8701_app0_stop_capture(&ctx);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(fakebuf[REG_APP0_CMD], APP0_CMD_STOP);
}

TEST_F (TMF8701Test, test_tof8701_app0_interrupt_status)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    uint8_t val = 0xAA;
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    rc = tof8701_app0_interrupt_status(NULL);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_interrupt_status(&ctx);
    ASSERT_EQ(rc, 0);
    read_idx = 0;
    *read_rc = -1;
    rc = tof8701_app0_interrupt_status(&ctx);
    ASSERT_LT(rc, 0);
    read_idx = 0;
    *read_rc = 0;
    fakebuf[REG_INT_STATUS] = val;
    rc = tof8701_app0_interrupt_status(&ctx);
    ASSERT_EQ(rc, val);
}

TEST_F (TMF8701Test, test_tof8701_app0_set_get_fac_calib)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    uint8_t in_vals[APP0_FAC_CALIB_SIZE] = {0};
    uint8_t out_vals[APP0_FAC_CALIB_SIZE] = {0};
    rc = tof8701_app0_set_fac_calib(NULL, in_vals, APP0_FAC_CALIB_SIZE);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_get_fac_calib(NULL, out_vals, APP0_FAC_CALIB_SIZE);
    ASSERT_EQ(rc, -1);
    for (int i = 0; i < APP0_FAC_CALIB_SIZE; i++) {
        in_vals[i] = i;
    }
    rc = tof8701_app0_set_fac_calib(&ctx, in_vals, APP0_FAC_CALIB_SIZE);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(ctx.app0_app.cal_flags & APP0_FAC_CAL_FLAG, APP0_FAC_CAL_FLAG);
    for (int i = 0; i < APP0_FAC_CALIB_SIZE; i++) {
        ASSERT_EQ(ctx.app0_app.app0_fac_calib[i], in_vals[i]);
    }
    rc = tof8701_app0_get_fac_calib(&ctx, out_vals, APP0_FAC_CALIB_SIZE);
    ASSERT_EQ(rc, 0);
    for (int i = 0; i < APP0_FAC_CALIB_SIZE; i++) {
        ASSERT_EQ(ctx.app0_app.app0_fac_calib[i], out_vals[i]);
    }
}

TEST_F (TMF8701Test, test_tof8701_app0_set_get_alg_state)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    uint8_t in_vals[APP0_ALG_STATE_SIZE] = {0};
    uint8_t out_vals[APP0_ALG_STATE_SIZE] = {0};
    rc = tof8701_app0_set_alg_state(NULL, in_vals, APP0_ALG_STATE_SIZE);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_get_alg_state(NULL, out_vals, APP0_ALG_STATE_SIZE);
    ASSERT_EQ(rc, -1);
    for (int i = 0; i < APP0_ALG_STATE_SIZE; i++) {
        in_vals[i] = i;
    }
    rc = tof8701_app0_set_alg_state(&ctx, in_vals, APP0_ALG_STATE_SIZE);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(ctx.app0_app.cal_flags & APP0_ALG_STATE_FLAG, APP0_ALG_STATE_FLAG);
    for (int i = 0; i < APP0_ALG_STATE_SIZE; i++) {
        ASSERT_EQ(ctx.app0_app.result.alg_state[i], in_vals[i]);
    }
    rc = tof8701_app0_get_alg_state(&ctx, out_vals, APP0_ALG_STATE_SIZE);
    ASSERT_EQ(rc, 0);
    for (int i = 0; i < APP0_ALG_STATE_SIZE; i++) {
        ASSERT_EQ(ctx.app0_app.result.alg_state[i], out_vals[i]);
    }
}

TEST_F (TMF8701Test, test_tof8701_app0_set_get_cfg_data)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    uint8_t in_vals[APP0_CFG_DATA_SIZE] = {0};
    uint8_t out_vals[APP0_CFG_DATA_SIZE] = {0};
    rc = tof8701_app0_set_cfg_data(NULL, in_vals, APP0_CFG_DATA_SIZE);
    ASSERT_EQ(rc, -1);
    rc = tof8701_app0_get_cfg_data(NULL, out_vals, APP0_CFG_DATA_SIZE);
    ASSERT_EQ(rc, -1);
    for (int i = 0; i < APP0_CFG_DATA_SIZE; i++) {
        in_vals[i] = i;
    }
    rc = tof8701_app0_set_cfg_data(&ctx, in_vals, APP0_CFG_DATA_SIZE);
    ASSERT_EQ(rc, 0);
    ASSERT_EQ(ctx.app0_app.cal_flags & APP0_CFG_DATA_FLAG, APP0_CFG_DATA_FLAG);
    for (int i = 0; i < APP0_CFG_DATA_SIZE; i++) {
        ASSERT_EQ(ctx.app0_app.app0_cfg_data[i], in_vals[i]);
    }
    rc = tof8701_app0_get_cfg_data(&ctx, out_vals, APP0_CFG_DATA_SIZE);
    ASSERT_EQ(rc, 0);
    for (int i = 0; i < APP0_CFG_DATA_SIZE; i++) {
        ASSERT_EQ(ctx.app0_app.app0_cfg_data[i], out_vals[i]);
    }
}

TEST_F (TMF8701Test, test_tof8701_app0_get_last_result)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    uint32_t dist = 0;
    uint32_t conf = 0;
    rc = tof8701_app0_get_last_result(NULL, &dist, &conf);
    ASSERT_LT(rc, 0);
    ctx.app0_app.result.distance_mm[0] = 1000 & 0xFF;
    ctx.app0_app.result.distance_mm[1] = 1000 >> 8;
    ctx.app0_app.result.info = 42;
    ctx.app0_app.result.res_num = 3;
    rc = tof8701_app0_get_last_result(&ctx, &dist, &conf);
    ASSERT_EQ(rc, 3);
    ASSERT_EQ(dist, 1000);
    ASSERT_EQ(conf, 42);
}

TEST_F (TMF8701Test, test_tof8701_app0_process_irq)
{
    int32_t rc = 0;
    struct tmf8701_ctx ctx = {0};
    uint32_t dist = 0;
    uint32_t conf = 0;
    tof8701_app0_process_irq(NULL);
    read_i2c_block_fake.custom_fake = fake_read;
    write_i2c_block_fake.custom_fake = fake_write;
    fakebuf[REG_INT_STATUS] = IRQ_RESULTS;
    fakebuf[REG_APP0_CONTENTS] = MEAS_RESULT_CONTENTS_ID;
    fakebuf[REG_APP0_DATA_START] = 42;
    tof8701_app0_process_irq(&ctx);
    ASSERT_EQ(tof8701_app0_get_last_result(&ctx, &dist, &conf), fakebuf[REG_APP0_DATA_START]);
    fakebuf[REG_INT_STATUS] = IRQ_RESULTS;
    fakebuf[REG_APP0_CONTENTS] = FAC_CALIB_CONTENTS_ID;
    tof8701_app0_process_irq(&ctx);
    ASSERT_EQ(fakebuf[REG_APP0_DATA_START], ctx.app0_app.app0_fac_calib[0]);
    fakebuf[REG_INT_STATUS] = IRQ_RESULTS;
    fakebuf[REG_APP0_CONTENTS] = 0xBB;
    tof8701_app0_process_irq(&ctx);
}
